from unittest.mock import patch

from django.test import override_settings

from app_utils.esi_testing import build_http_error
from app_utils.testing import (
    NoSocketsTestCase,
    create_user_from_evecharacter,
    reset_celery_once_locks,
)

from blueprints import tasks

from . import create_owner
from .testdata.load_entities import load_entities
from .testdata.load_eveuniverse import load_eveuniverse
from .testdata.load_locations import load_locations

TASKS_PATH = "blueprints.tasks"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestTasks(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        load_locations()
        cls.owner = create_owner(character_id=1101, corporation_id=2101)

    @patch(TASKS_PATH + ".Owner.update_blueprints_esi")
    def test_update_all_blueprints(self, mock_update_blueprints_esi):
        tasks.update_all_blueprints()
        self.assertTrue(mock_update_blueprints_esi.called)

    @patch(TASKS_PATH + ".Owner.update_blueprints_esi")
    def test_update_blueprints_for_owner(self, mock_update_blueprints_esi):
        tasks.update_blueprints_for_owner(self.owner.pk)
        self.assertTrue(mock_update_blueprints_esi.called)

    @patch(TASKS_PATH + ".Owner.update_industry_jobs_esi")
    def test_update_all_industry_jobs(self, mock_update_industry_jobs_esi):
        tasks.update_all_industry_jobs()
        self.assertTrue(mock_update_industry_jobs_esi.called)

    @patch(TASKS_PATH + ".Owner.update_industry_jobs_esi")
    def test_update_industry_job_for_owner(self, mock_update_industry_jobs_esi):
        tasks.update_industry_jobs_for_owner(self.owner.pk)
        self.assertTrue(mock_update_industry_jobs_esi.called)

    @patch(TASKS_PATH + ".Owner.update_locations_esi")
    def test_update_all_locations(self, mock_update_locations_esi):
        tasks.update_all_locations()
        self.assertTrue(mock_update_locations_esi.called)

    @patch(TASKS_PATH + ".Owner.update_locations_esi")
    def test_update_locations_for_owner(self, mock_update_locations_esi):
        tasks.update_locations_for_owner(self.owner.pk)
        self.assertTrue(mock_update_locations_esi.called)


@patch("celery.app.task.Context.called_directly", False)  # make retry work with eager
@override_settings(CELERY_ALWAYS_EAGER=True)
@patch(TASKS_PATH + ".Location.objects.structure_update_or_create_esi")
class TestUpdateStructures(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.user, _ = create_user_from_evecharacter(1001)
        cls.token = cls.user.token_set.first()

    def setUp(self):
        reset_celery_once_locks("blueprints")

    def test_should_update_structure(self, mock_structure_update_or_create_esi):
        # when
        tasks.update_structure_esi.delay(
            structure_id=1000000000001, token_pk=self.token.pk
        )

        # then
        self.assertEqual(mock_structure_update_or_create_esi.call_count, 1)

    def test_should_retry_when_esi_is_offline(
        self, mock_structure_update_or_create_esi
    ):
        # given
        mock_structure_update_or_create_esi.side_effect = [
            build_http_error(502),
            lambda: None,
        ]
        # when
        tasks.update_structure_esi.delay(
            structure_id=1000000000001, token_pk=self.token.pk
        )
        # then
        self.assertEqual(mock_structure_update_or_create_esi.call_count, 2)

    def test_should_abort_on_other_exceptions(
        self, mock_structure_update_or_create_esi
    ):
        # given
        mock_structure_update_or_create_esi.side_effect = [
            build_http_error(500),
            lambda: None,
        ]
        # when
        tasks.update_structure_esi.delay(
            structure_id=1000000000001, token_pk=self.token.pk
        )
        # then
        self.assertEqual(mock_structure_update_or_create_esi.call_count, 1)



from django.core.management.base import BaseCommand

from blueprints.tasks import update_all_blueprints, update_all_locations


class Command(BaseCommand):
    help = "Refreshes all blueprints and locations"

    def handle(self, *args, **options):
        update_all_blueprints.delay()
        update_all_locations.delay()
        self.stdout.write("Blueprint and location refresh tasks enqueued.")



"""Views for rending blueprint lists for Blueprints."""

from typing import Any

from dj_datatables_view.base_datatable_view import BaseDatatableView

from django.utils.html import format_html
from django.utils.translation import gettext_lazy
from eveuniverse.models import EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from blueprints import __title__
from blueprints.app_settings import BLUEPRINTS_LIST_ICON_OUTPUT_SIZE
from blueprints.models import Blueprint

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


# pylint: disable = too-many-ancestors
class BlueprintListJson(BaseDatatableView):
    """View for creating a list view of blueprints for datatables."""

    # The model we're going to show
    model = Blueprint

    # define the columns that will be returned
    columns = [
        "eve_type_icon",
        "eve_type",
        "quantity",
        "owner",
        "material_efficiency",
        "time_efficiency",
        "is_original",
        "runs",
        "pk",
        "location",
        "industryjob",
    ]

    # define column names that will be used in sorting
    # order is important and should be same as order of columns
    # displayed by datatables. For non sortable columns use empty
    # value like ''
    order_columns = [
        "eve_type_icon",
        "eve_type",
        "",
        "owner",
        "material_efficiency",
        "time_efficiency",
        "is_original",
        "runs",
        "pk",
        "location",
        "industryjob",
    ]

    # set max limit of records returned, this is used to protect our site if someone tries to attack our site
    # and make it return huge amount of data
    max_display_length = 500

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._user_has_location_permission = None

    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)
        self._user_has_location_permission = self.request.user.has_perm(
            "blueprints.view_blueprint_locations"
        )

    def get_initial_queryset(self):
        """Return queryset used as base for further sorting/filtering

        these are simply objects displayed in datatable
        You should not filter data returned here by any filter values entered by user. This is because
        we need some base queryset to count total number of records.
        """
        return Blueprint.objects.user_has_access(
            self.request.user
        ).annotate_location_name()

    def filter_queryset(self, qs):
        def apply_search_filter(qs, column_num, field):
            my_filter = self.request.GET.get(
                f"columns[{column_num}][search][value]", None
            )
            if my_filter:
                if self.request.GET.get(f"columns[{column_num}][search][regex]", False):
                    kwargs = {f"{field}__iregex": my_filter}
                else:
                    kwargs = {f"{field}__istartswith": my_filter}
                return qs.filter(**kwargs)
            return qs

        qs = qs.annotate_is_bpo().annotate_owner_name()
        qs = apply_search_filter(qs, 9, "location_name")
        qs = apply_search_filter(qs, 3, "owner_name")
        qs = apply_search_filter(qs, 4, "material_efficiency")
        qs = apply_search_filter(qs, 5, "time_efficiency")
        qs = apply_search_filter(qs, 6, "is_bpo")

        search = self.request.GET.get("search[value]", None)
        if search:
            qs = qs.filter(eve_type__name__icontains=search)

        return qs

    # pylint: disable = too-many-return-statements
    def render_column(self, row, column):
        if column == "eve_type_icon":
            variant = EveType.IconVariant.BPC if row.runs else EveType.IconVariant.BPO
            return format_html(
                '<img src="{}" width="{}" height="{}">',
                row.eve_type.icon_url(size=64, variant=variant),
                BLUEPRINTS_LIST_ICON_OUTPUT_SIZE,
                BLUEPRINTS_LIST_ICON_OUTPUT_SIZE,
            )

        if column == "location":
            if self._user_has_location_permission:
                return row.location_name
            return gettext_lazy("(Unknown)")

        if column == "is_original":
            return "Yes" if row.is_original else "No"

        if column == "owner":
            if row.owner.corporation:
                return {
                    "display": (
                        "<span class='fas fa-briefcase'></span>&nbsp;"
                        + row.owner.corporation.corporation_name
                    ),
                    "sort": row.owner.corporation.corporation_name,
                }
            return {
                "display": (
                    "<span class='fas fa-user'></span>&nbsp;"
                    + row.owner.character.character.character_name
                ),
                "sort": row.owner.character.character.character_name,
            }

        return super().render_column(row, column)


# @login_required
# @permissions_required("blueprints.basic_access")
# def list_blueprints(request):
#     from . import convert_blueprint

#     blueprint_rows = [
#         convert_blueprint(blueprint, request.user) for blueprint in blueprints_query
#     ]

#     return JsonResponse(blueprint_rows, safe=False)



"""Models for Blueprints."""

from django.contrib.auth.models import Permission, User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from esi.errors import TokenError
from esi.models import Token
from eveuniverse.models import EveEntity, EveSolarSystem, EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.evelinks import dotlan
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.django import users_with_permission
from app_utils.logging import LoggerAddTag

from . import __title__
from .managers import BlueprintManager, LocationManager, OwnerManager, RequestManager
from .providers import esi
from .validators import validate_material_efficiency, validate_time_efficiency

NAMES_MAX_LENGTH = 100

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def get_or_create_location_async(location_id: int, token: Token) -> "Location":
    """Get or create a location sync - helper function."""
    obj, _ = Location.objects.get_or_create_esi_async(id=location_id, token=token)
    return obj


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app"),
            ("request_blueprints", "Can request blueprints"),
            ("manage_requests", "Can review and accept blueprint requests"),
            ("add_personal_blueprint_owner", "Can add personal blueprint owners"),
            ("add_corporate_blueprint_owner", "Can add corporate blueprint owners"),
            ("view_alliance_blueprints", "Can view alliance's blueprints"),
            ("view_blueprint_locations", "Can view the location of all blueprints"),
            ("view_industry_jobs", "Can view details about running industry jobs"),
        )


class Owner(models.Model):
    """A corporation that owns blueprints"""

    corporation = models.OneToOneField(
        EveCorporationInfo,
        default=None,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        help_text="Corporation owning blueprints, if this is a 'corporate' owner",
        related_name="+",
    )
    character = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        help_text="character used for syncing blueprints",
        related_name="+",
    )
    is_active = models.BooleanField(
        default=True,
        help_text=("whether this owner is currently included in the sync process"),
    )

    objects = OwnerManager()

    class Meta:
        default_permissions = ()

    def __str__(self):
        return self.name

    @property
    def name(self) -> str:
        """Return the name of this owner."""
        try:
            if self.corporation:
                return self.corporation.corporation_name
            return self.eve_character_strict.character_name
        except (ValueError, AttributeError):
            return ""

    @property
    def corporation_strict(self) -> EveCorporationInfo:
        """Return corporation of this owner when it exists, or raises error."""
        if not self.corporation:
            raise ValueError("No corporation defined")
        return self.corporation

    @property
    def eve_character_strict(self) -> EveCharacter:
        """Return character of this owner when it exists, or raises error."""
        if not self.character or not self.character.character:
            raise ValueError("No character defined")
        return self.character.character

    def update_locations_esi(self):
        """Update locations from ESI."""
        if self.corporation:
            token = self.valid_token(
                [
                    "esi-universe.read_structures.v1",
                    "esi-assets.read_corporation_assets.v1",
                ]
            )
            assets = self._fetch_corporate_assets(token)
        else:
            token = self.valid_token(
                ["esi-universe.read_structures.v1", "esi-assets.read_assets.v1"]
            )
            assets = self._fetch_personal_assets(token)

        asset_ids = []
        asset_locations = {}
        assets_by_id = {}
        for asset in assets:
            asset_ids.append(asset["item_id"])
            assets_by_id[asset["item_id"]] = asset

        for asset in assets:
            if asset["location_id"] in asset_ids:
                location_id = asset["location_id"]
                if asset_locations.get(location_id):
                    asset_locations[location_id].append(asset["item_id"])
                else:
                    asset_locations[location_id] = [asset["item_id"]]

        for location_id in asset_locations:
            asset = assets_by_id[location_id]
            parent_location = asset["location_id"]
            parent = get_or_create_location_async(parent_location, token=token)
            eve_type, _ = EveType.objects.get_or_create_esi(id=asset["type_id"])
            Location.objects.update_or_create(
                id=location_id, defaults={"parent": parent, "eve_type": eve_type}
            )

    def update_blueprints_esi(self):
        """Update all blueprints from ESI."""

        blueprint_ids_to_remove = list(
            self.blueprints.values_list("item_id", flat=True)
        )
        if self.corporation:
            token = self.valid_token(
                [
                    "esi-universe.read_structures.v1",
                    "esi-corporations.read_blueprints.v1",
                ]
            )
            blueprints = self._fetch_corporate_blueprints(token)
        else:
            token = self.valid_token(
                [
                    "esi-universe.read_structures.v1",
                    "esi-characters.read_blueprints.v1",
                ]
            )
            blueprints = self._fetch_personal_blueprints(token)

        for blueprint in blueprints:
            runs = blueprint["runs"]
            if runs < 1:
                runs = None
            quantity = blueprint["quantity"]
            if quantity < 0:
                quantity = 1
            original = self.blueprints.filter(item_id=blueprint["item_id"]).first()

            location_flag = Blueprint.LocationFlag.from_esi_data(
                blueprint["location_flag"]
            )
            eve_type, _ = EveType.objects.get_or_create_esi(id=blueprint["type_id"])
            if original is not None:
                # We've seen this blueprint coming from ESI, so we know it shouldn't be deleted
                blueprint_ids_to_remove.remove(original.item_id)
                original.location = get_or_create_location_async(
                    blueprint["location_id"],
                    token=token,
                )
                original.location_flag = location_flag
                original.eve_type = eve_type
                original.runs = runs
                original.material_efficiency = blueprint["material_efficiency"]
                original.time_efficiency = blueprint["time_efficiency"]
                original.quantity = quantity
                original.save()
            else:
                self.blueprints.create(
                    location=get_or_create_location_async(
                        blueprint["location_id"], token=token
                    ),
                    location_flag=location_flag,
                    eve_type=eve_type,
                    item_id=blueprint["item_id"],
                    runs=runs,
                    material_efficiency=blueprint["material_efficiency"],
                    time_efficiency=blueprint["time_efficiency"],
                    quantity=quantity,
                )

        Blueprint.objects.filter(pk__in=blueprint_ids_to_remove).delete()

    def update_industry_jobs_esi(self):
        """Update all blueprints from ESI."""

        if not self.is_active:
            return

        job_ids_to_remove = list(
            IndustryJob.objects.filter(owner=self).values_list("id", flat=True)
        )
        if self.corporation:
            token = self.valid_token(
                [
                    "esi-universe.read_structures.v1",
                    "esi-industry.read_corporation_jobs.v1",
                ]
            )
            jobs = self._fetch_corporate_industry_jobs(token)
        else:
            token = self.valid_token(
                [
                    "esi-universe.read_structures.v1",
                    "esi-industry.read_character_jobs.v1",
                ]
            )
            jobs = self._fetch_personal_industry_jobs(token)

        for job in jobs:
            original = IndustryJob.objects.filter(owner=self, id=job["job_id"]).first()
            blueprint = Blueprint.objects.filter(pk=job["blueprint_id"]).first()
            if blueprint is not None:
                if original is not None:
                    # We've seen this job coming from ESI, so we know it shouldn't be deleted
                    job_ids_to_remove.remove(original.id)
                    original.status = job["status"]
                    original.save()
                else:
                    # Reject personal listings of corporate jobs and visa-versa
                    if blueprint.owner == self:
                        installer = EveCharacter.objects.get_character_by_id(
                            job["installer_id"]
                        )
                        if not installer:
                            installer = EveCharacter.objects.create_character(
                                job["installer_id"]
                            )
                        IndustryJob.objects.create(
                            id=job["job_id"],
                            activity=job["activity_id"],
                            owner=self,
                            location=get_or_create_location_async(
                                job["output_location_id"],
                                token=token,
                            ),
                            blueprint=Blueprint.objects.get(pk=job["blueprint_id"]),
                            installer=installer,
                            runs=job["runs"],
                            start_date=job["start_date"],
                            end_date=job["end_date"],
                            status=job["status"],
                        )
            else:
                blueprint_id = job["blueprint_id"]
                logger.warning(f"Unmatchable blueprint ID: {blueprint_id}")

        IndustryJob.objects.filter(pk__in=job_ids_to_remove).delete()

    def _fetch_corporate_assets(self, token) -> list:
        return esi.client.Assets.get_corporations_corporation_id_assets(
            corporation_id=self.corporation_strict.corporation_id,
            token=token.valid_access_token(),
        ).results()

    def _fetch_personal_assets(self, token) -> list:
        return esi.client.Assets.get_characters_character_id_assets(
            character_id=self.eve_character_strict.character_id,
            token=token.valid_access_token(),
        ).results()

    def _fetch_corporate_blueprints(self, token: Token) -> list:
        blueprints = esi.client.Corporation.get_corporations_corporation_id_blueprints(
            corporation_id=self.corporation_strict.corporation_id,
            token=token.valid_access_token(),
        ).results()
        return blueprints

    def _fetch_personal_blueprints(self, token: Token) -> list:
        blueprints = esi.client.Character.get_characters_character_id_blueprints(
            character_id=self.eve_character_strict.character_id,
            token=token.valid_access_token(),
        ).results()
        return blueprints

    def _fetch_corporate_industry_jobs(self, token: Token) -> list:
        jobs = esi.client.Industry.get_corporations_corporation_id_industry_jobs(
            corporation_id=self.corporation_strict.corporation_id,
            token=token.valid_access_token(),
        ).results()
        return jobs

    def _fetch_personal_industry_jobs(self, token: Token) -> list:
        jobs = esi.client.Industry.get_characters_character_id_industry_jobs(
            character_id=self.eve_character_strict.character_id,
            token=token.valid_access_token(),
        ).results()
        return jobs

    def valid_token(self, scopes) -> Token:
        """Return a valid token for the owner or raise exception."""
        token = (
            Token.objects.filter(
                user=self.character.user,
                character_id=self.eve_character_strict.character_id,
            )
            .require_scopes(scopes)
            .require_valid()
            .first()
        )
        if not token:
            raise TokenError(f"{self}: No valid token found with sufficient scopes")

        return token


class Blueprint(models.Model):
    """A blueprint in Eve Online."""

    class LocationFlag(models.TextChoices):
        """A flag denoting the location of a blueprint."""

        ASSET_SAFETY = "AssetSafety", _("Asset Safety")
        AUTO_FIT = "AutoFit", _("Auto Fit")
        BONUS = "Bonus", _("Bonus")
        BOOSTER = "Booster", _("Booster")
        BOOSTER_BAY = "BoosterBay", _("Booster Hold")
        CAPSULE = "Capsule", _("Capsule")
        CARGO = "Cargo", _("Cargo")
        CORP_DELIVERIES = "CorpDeliveries", _("Corp Deliveries")
        CORP_S_A_G_1 = "CorpSAG1", _("Corp Security Access Group 1")
        CORP_S_A_G_2 = "CorpSAG2", _("Corp Security Access Group 2")
        CORP_S_A_G_3 = "CorpSAG3", _("Corp Security Access Group 3")
        CORP_S_A_G_4 = "CorpSAG4", _("Corp Security Access Group 4")
        CORP_S_A_G_5 = "CorpSAG5", _("Corp Security Access Group 5")
        CORP_S_A_G_6 = "CorpSAG6", _("Corp Security Access Group 6")
        CORP_S_A_G_7 = "CorpSAG7", _("Corp Security Access Group 7")
        CRATE_LOOT = "CrateLoot", _("Crate Loot")
        DELIVERIES = "Deliveries", _("Deliveries")
        DRONE_BAY = "DroneBay", _("Drone Bay")
        DUST_BATTLE = "DustBattle", _("Dust Battle")
        DUST_DATABANK = "DustDatabank", _("Dust Databank")
        FIGHTER_BAY = "FighterBay", _("Fighter Bay")
        FIGHTER_TUBE_0 = "FighterTube0", _("Fighter Tube 0")
        FIGHTER_TUBE_1 = "FighterTube1", _("Fighter Tube 1")
        FIGHTER_TUBE_2 = "FighterTube2", _("Fighter Tube 2")
        FIGHTER_TUBE_3 = "FighterTube3", _("Fighter Tube 3")
        FIGHTER_TUBE_4 = "FighterTube4", _("Fighter Tube 4")
        FLEET_HANGAR = "FleetHangar", _("Fleet Hangar")
        FRIGATE_ESCAPE_BAY = "FrigateEscapeBay", _("Frigate escape bay Hangar")
        HANGAR = "Hangar", _("Hangar")
        HANGAR_ALL = "HangarAll", _("Hangar All")
        HI_SLOT_0 = "HiSlot0", _("High power slot 1")
        HI_SLOT_1 = "HiSlot1", _("High power slot 2")
        HI_SLOT_2 = "HiSlot2", _("High power slot 3")
        HI_SLOT_3 = "HiSlot3", _("High power slot 4")
        HI_SLOT_4 = "HiSlot4", _("High power slot 5")
        HI_SLOT_5 = "HiSlot5", _("High power slot 6")
        HI_SLOT_6 = "HiSlot6", _("High power slot 7")
        HI_SLOT_7 = "HiSlot7", _("High power slot 8")
        HIDDEN_MODIFIERS = "HiddenModifiers", _("Hidden Modifiers")
        IMPLANT = "Implant", _("Implant")
        IMPOUNDED = "Impounded", _("Impounded")
        JUNKYARD_REPROCESSED = "JunkyardReprocessed", _(
            "This item was put into a junkyard through reprocessing."
        )
        JUNKYARD_TRASHED = "JunkyardTrashed", _(
            "This item was put into a junkyard through being trashed by its owner."
        )
        LO_SLOT_0 = "LoSlot0", _("Low power slot 1")
        LO_SLOT_1 = "LoSlot1", _("Low power slot 2")
        LO_SLOT_2 = "LoSlot2", _("Low power slot 3")
        LO_SLOT_3 = "LoSlot3", _("Low power slot 4")
        LO_SLOT_4 = "LoSlot4", _("Low power slot 5")
        LO_SLOT_5 = "LoSlot5", _("Low power slot 6")
        LO_SLOT_6 = "LoSlot6", _("Low power slot 7")
        LO_SLOT_7 = "LoSlot7", _("Low power slot 8")
        LOCKED = "Locked", _("Locked item, can not be moved unless unlocked")
        MED_SLOT_0 = "MedSlot0", _("Medium power slot 1")
        MED_SLOT_1 = "MedSlot1", _("Medium power slot 2")
        MED_SLOT_2 = "MedSlot2", _("Medium power slot 3")
        MED_SLOT_3 = "MedSlot3", _("Medium power slot 4")
        MED_SLOT_4 = "MedSlot4", _("Medium power slot 5")
        MED_SLOT_5 = "MedSlot5", _("Medium power slot 6")
        MED_SLOT_6 = "MedSlot6", _("Medium power slot 7")
        MED_SLOT_7 = "MedSlot7", _("Medium power slot 8")
        OFFICE_FOLDER = "OfficeFolder", _("Office Folder")
        PILOT = "Pilot", _("Pilot")
        PLANET_SURFACE = "PlanetSurface", _("Planet Surface")
        QUAFE_BAY = "QuafeBay", _("Quafe Bay")
        QUANTUM_CORE_ROOM = "QuantumCoreRoom", _("Quantum Core Room")
        REWARD = "Reward", _("Reward")
        RIG_SLOT_0 = "RigSlot0", _("Rig power slot 1")
        RIG_SLOT_1 = "RigSlot1", _("Rig power slot 2")
        RIG_SLOT_2 = "RigSlot2", _("Rig power slot 3")
        RIG_SLOT_3 = "RigSlot3", _("Rig power slot 4")
        RIG_SLOT_4 = "RigSlot4", _("Rig power slot 5")
        RIG_SLOT_5 = "RigSlot5", _("Rig power slot 6")
        RIG_SLOT_6 = "RigSlot6", _("Rig power slot 7")
        RIG_SLOT_7 = "RigSlot7", _("Rig power slot 8")
        SECONDARY_STORAGE = "SecondaryStorage", _("Secondary Storage")
        SERVICE_SLOT_0 = "ServiceSlot0", _("Service Slot 0")
        SERVICE_SLOT_1 = "ServiceSlot1", _("Service Slot 1")
        SERVICE_SLOT_2 = "ServiceSlot2", _("Service Slot 2")
        SERVICE_SLOT_3 = "ServiceSlot3", _("Service Slot 3")
        SERVICE_SLOT_4 = "ServiceSlot4", _("Service Slot 4")
        SERVICE_SLOT_5 = "ServiceSlot5", _("Service Slot 5")
        SERVICE_SLOT_6 = "ServiceSlot6", _("Service Slot 6")
        SERVICE_SLOT_7 = "ServiceSlot7", _("Service Slot 7")
        SHIP_HANGAR = "ShipHangar", _("Ship Hangar")
        SHIP_OFFLINE = "ShipOffline", _("Ship Offline")
        SKILL = "Skill", _("Skill")
        SKILL_IN_TRAINING = "SkillInTraining", _("Skill In Training")
        SPECIALIZED_AMMO_HOLD = "SpecializedAmmoHold", _("Specialized Ammo Hold")
        SPECIALIZED_COMMAND_CENTER_HOLD = "SpecializedCommandCenterHold", _(
            "Specialized Command Center Hold"
        )
        SPECIALIZED_FUEL_BAY = "SpecializedFuelBay", _("Specialized Fuel Bay")
        SPECIALIZED_GAS_HOLD = "SpecializedGasHold", _("Specialized Gas Hold")
        SPECIALIZED_INDUSTRIAL_SHIP_HOLD = "SpecializedIndustrialShipHold", _(
            "Specialized Industrial Ship Hold"
        )
        SPECIALIZED_LARGE_SHIP_HOLD = "SpecializedLargeShipHold", _(
            "Specialized Large Ship Hold"
        )
        SPECIALIZED_MATERIAL_BAY = "SpecializedMaterialBay", _(
            "Specialized Material Bay"
        )
        SPECIALIZED_MEDIUM_SHIP_HOLD = "SpecializedMediumShipHold", _(
            "Specialized Medium Ship Hold"
        )
        SPECIALIZED_MINERAL_HOLD = "SpecializedMineralHold", _(
            "Specialized Mineral Hold"
        )
        SPECIALIZED_ORE_HOLD = "SpecializedOreHold", _("Specialized Ore Hold")
        SPECIALIZED_PLANETARY_COMMODITIES_HOLD = (
            "SpecializedPlanetaryCommoditiesHold",
            _("Specialized Planetary Commodities Hold"),
        )
        SPECIALIZED_SALVAGE_HOLD = "SpecializedSalvageHold", _(
            "Specialized Salvage Hold"
        )
        SPECIALIZED_SHIP_HOLD = "SpecializedShipHold", _("Specialized Ship Hold")
        SPECIALIZED_SMALL_SHIP_HOLD = "SpecializedSmallShipHold", _(
            "Specialized Small Ship Hold"
        )
        STRUCTURE_ACTIVE = "StructureActive", _("Structure Active")
        STRUCTURE_FUEL = "StructureFuel", _("Structure Fuel")
        STRUCTURE_INACTIVE = "StructureInactive", _("Structure Inactive")
        STRUCTURE_OFFLINE = "StructureOffline", _("Structure Offline")
        SUB_SYSTEM_BAY = "SubSystemBay", _("Sub System Bay")
        SUB_SYSTEM_SLOT_0 = "SubSystemSlot0", _("Sub System Slot 0")
        SUB_SYSTEM_SLOT_1 = "SubSystemSlot1", _("Sub System Slot 1")
        SUB_SYSTEM_SLOT_2 = "SubSystemSlot2", _("Sub System Slot 2")
        SUB_SYSTEM_SLOT_3 = "SubSystemSlot3", _("Sub System Slot 3")
        SUB_SYSTEM_SLOT_4 = "SubSystemSlot4", _("Sub System Slot 4")
        SUB_SYSTEM_SLOT_5 = "SubSystemSlot5", _("Sub System Slot 5")
        SUB_SYSTEM_SLOT_6 = "SubSystemSlot6", _("Sub System Slot 6")
        SUB_SYSTEM_SLOT_7 = "SubSystemSlot7", _("Sub System Slot 7")
        UNLOCKED = "Unlocked", _("Unlocked item, can be moved")
        WALLET = "Wallet", _("Wallet")
        WARDROBE = "Wardrobe", _("Wardrobe")
        UNDEFINED = "Undefined", _("undefined")

        @classmethod
        def from_esi_data(cls, data: str) -> "Blueprint.LocationFlag":
            """Create new obj from ESI data."""
            try:
                return cls(data)
            except ValueError:
                return cls.UNDEFINED

    item_id = models.PositiveBigIntegerField(
        primary_key=True, help_text="The EVE Item ID of the blueprint"
    )
    owner = models.ForeignKey(
        Owner,
        on_delete=models.CASCADE,
        related_name="blueprints",
        help_text="Corporation that owns the blueprint",
    )
    eve_type = models.ForeignKey(
        EveType, on_delete=models.CASCADE, related_name="+", help_text="Blueprint type"
    )
    location = models.ForeignKey(
        "Location",
        on_delete=models.CASCADE,
        related_name="blueprints",
        help_text="Blueprint location",
    )
    location_flag = models.CharField(
        help_text="Additional location information",
        choices=LocationFlag.choices,
        max_length=36,
    )
    quantity = models.PositiveIntegerField(help_text="Number of blueprints", default=1)
    runs = models.PositiveIntegerField(
        blank=True,
        null=True,
        help_text="Runs remaining or null if the blueprint is an original",
    )
    material_efficiency = models.PositiveIntegerField(
        help_text="Material efficiency of the blueprint",
        validators=[validate_material_efficiency],
    )
    time_efficiency = models.PositiveIntegerField(
        help_text="Time efficiency of the blueprint",
        validators=[validate_time_efficiency],
    )

    objects = BlueprintManager()

    @property
    def is_original(self):
        """Return True, when this is a BPO, else False"""
        return not self.runs

    @property
    def location_flag_obj(self) -> "Blueprint.LocationFlag":
        """Return the location flag object of this blueprint."""
        return self.LocationFlag(self.location_flag)

    class Meta:
        default_permissions = ()

    def __str__(self):
        return (
            self.eve_type.name + f" ({self.material_efficiency}/{self.time_efficiency})"
        )

    def has_industryjob(self):
        """Return True if this blueprint has an industry job, else False."""
        try:
            return self.industryjob is not None  # pylint: disable = no-member
        except ObjectDoesNotExist:
            return False


class IndustryJob(models.Model):
    """An industry job in Eve Online."""

    id = models.PositiveBigIntegerField(
        primary_key=True,
        help_text=("Eve Online job ID"),
    )

    class Activity(models.IntegerChoices):
        """The type of activity an industry job is performing."""

        MANUFACTURING = 1, _("Manufacturing")
        RESEARCHING_TECHNOLOGY = 2, _("Researching Technology")
        RESEARCHING_TIME_EFFICIENCY = 3, _("Researching Time Efficiency")
        RESEARCHING_MATERIAL_EFFICIENCY = 4, _("Researching Material Efficiency")
        COPYING = 5, _("Copying")
        DUPLICATING = 6, _("Duplicating")
        REVERSE_ENGINEERING = 7, _("Reverse Engineering")
        INVENTING = 8, _("Inventing")
        REACTING = 9, _("Reacting")

    activity = models.PositiveIntegerField(choices=Activity.choices)
    location = models.ForeignKey(
        "Location",
        on_delete=models.CASCADE,
        help_text=(
            "Eve Online location ID of the facility in which the job is running"
        ),
        related_name="+",
    )
    blueprint = models.OneToOneField(
        Blueprint,
        on_delete=models.CASCADE,
        help_text=("Blueprint the job is running"),
    )
    installer = models.ForeignKey(
        EveCharacter,
        on_delete=models.CASCADE,
        related_name="jobs",
    )
    owner = models.ForeignKey(
        Owner,
        on_delete=models.CASCADE,
        related_name="jobs",
    )
    start_date = models.DateTimeField()
    end_date = models.DateTimeField()
    runs = models.PositiveIntegerField()
    status = models.CharField(
        choices=(
            ("active", "Active"),
            ("cancelled", "Cancelled"),
            ("delivered", "Delivered"),
            ("paused", "Paused"),
            ("ready", "Ready"),
            ("reverted", "Reverted"),
        ),
        max_length=10,
    )


class Location(models.Model):
    """An Eve Online location: Station or Upwell Structure or Solar System"""

    _SOLAR_SYSTEM_ID_START = 30_000_000
    _SOLAR_SYSTEM_ID_END = 33_000_000
    _STATION_ID_START = 60_000_000
    _STATION_ID_END = 64_000_000

    id = models.PositiveBigIntegerField(
        primary_key=True,
        help_text=(
            "Eve Online location ID, "
            "either item ID for stations or structure ID for structures"
        ),
    )
    parent = models.ForeignKey(
        "Location",
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        help_text=("Eve Online location ID of the parent object"),
        related_name="+",
    )

    name = models.CharField(
        max_length=NAMES_MAX_LENGTH,
        help_text="In-game name of this station or structure",
    )
    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    owner = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        help_text="corporation this station or structure belongs to",
        related_name="+",
    )
    updated_at = models.DateTimeField(auto_now=True)

    objects = LocationManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        if self.name:
            return self.name
        if self.eve_type:
            return str(self.eve_type)
        return f"Location #{self.id}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name='{self.name}')"

    @property
    def is_empty(self) -> bool:
        """Return True if this is an empty location, else False."""
        return not self.eve_solar_system and not self.eve_type and not self.parent_id

    @property
    def solar_system_url(self) -> str:
        """Return dotlan URL for this solar system."""
        try:
            return dotlan.solar_system_url(self.eve_solar_system.name)
        except AttributeError:
            return ""

    @property
    def is_solar_system(self) -> bool:
        """Return True if this location is a solar system, else False."""
        return self.is_solar_system_id(self.id)

    @property
    def is_station(self) -> bool:
        """Return True if this location is a station, else False."""
        return self.is_station_id(self.id)

    @classmethod
    def is_solar_system_id(cls, location_id: int) -> bool:
        """Return True if the given ID is a solar system ID, else False."""
        return cls._SOLAR_SYSTEM_ID_START <= location_id <= cls._SOLAR_SYSTEM_ID_END

    @classmethod
    def is_station_id(cls, location_id: int) -> bool:
        """Return True if the given ID is a station ID, else False."""
        return cls._STATION_ID_START <= location_id <= cls._STATION_ID_END

    def full_qualified_name(self) -> str:
        """Return the full qualified name of this location."""
        if self.parent:
            return f"{self.parent.full_qualified_name()} - {str(self)}"
        return str(self)


class Request(models.Model):
    """A request to use a specific blueprint."""

    blueprint = models.ForeignKey(
        Blueprint,
        on_delete=models.CASCADE,
    )
    requesting_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name="+",
        help_text="The requesting user",
    )
    fulfulling_user = models.ForeignKey(
        User,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="+",
        help_text="The user that fulfilled the request, if it has been fulfilled",
    )
    runs = models.PositiveIntegerField(
        blank=True,
        null=True,
        help_text="Runs requested or blank for maximum allowed",
    )
    STATUS_OPEN = "OP"
    STATUS_IN_PROGRESS = "IP"
    STATUS_FULFILLED = "FL"
    STATUS_CANCELLED = "CL"

    STATUS_CHOICES = [
        (STATUS_OPEN, "Open"),
        (STATUS_IN_PROGRESS, "In Progress"),
        (STATUS_FULFILLED, "Fulfilled"),
        (STATUS_CANCELLED, "Cancelled"),
    ]
    status = models.CharField(
        help_text="Status of the blueprint request",
        choices=STATUS_CHOICES,
        max_length=2,
        db_index=True,
    )
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)
    closed_at = models.DateTimeField(blank=True, null=True, db_index=True)

    objects = RequestManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        character_name = self.requesting_character_name()
        type_name = self.blueprint.eve_type.name
        return f"{character_name}'s request for {type_name}"

    def __repr__(self) -> str:
        character_name = self.requesting_character_name()
        return (
            f"{self.__class__.__name__}(id={self.pk}, "
            f"requesting_user='{character_name}', "
            f"type_name='{self.blueprint.eve_type.name}')"
        )

    def requesting_character_name(self) -> str:
        """Return main character's name of the requesting user safely."""
        try:
            return self.requesting_user.profile.main_character.character_name
        except AttributeError:
            return "?"

    def mark_request(
        self, user: User, status: str, closed: bool, can_requestor_edit: bool = False
    ) -> bool:
        """Change the status of a blueprint request."""
        character_ownerships = user.character_ownerships.select_related(
            "character"
        ).all()
        corporation_ids = {
            character.character.corporation_id for character in character_ownerships
        }
        character_ownership_ids = {character.pk for character in character_ownerships}
        has_requestor_character_in_owner_corporation = (
            self.blueprint.owner.corporation
            and self.blueprint.owner.corporation.corporation_id in corporation_ids
        )
        is_requestor_owner_of_blueprint = (
            not self.blueprint.owner.corporation
            and self.blueprint.owner.character
            and self.blueprint.owner.character.pk in character_ownership_ids
        )

        if (
            has_requestor_character_in_owner_corporation
            or is_requestor_owner_of_blueprint
            or (can_requestor_edit and self.requesting_user == user)
        ):
            if closed:
                self.closed_at = now()
            else:
                self.closed_at = None

            if status in {Request.STATUS_FULFILLED, Request.STATUS_IN_PROGRESS}:
                fulfulling_user = user
            else:
                fulfulling_user = None

            self.fulfulling_user = fulfulling_user
            self.status = status
            self.save()
            return True

        return False

    def notify_new_request(self) -> None:
        """Notify approvers that a blueprint request has been created."""
        for approver in self.approvers():
            notify(
                title=(
                    f"Blueprints: {self.blueprint.eve_type.name} has been requested"
                ),
                message=(
                    f"A copy of {self.blueprint.eve_type.name} has been "
                    f"requested by {self.requesting_user}."
                ),
                user=approver,
                level="info",
            )

    def notify_request_in_progress(self) -> None:
        """Notify requestor and approvers that a blueprint request is in progress."""
        notify(
            title=(f"Blueprints: {self.blueprint.eve_type.name} request in progress"),
            message=(
                f"{self.fulfulling_user} has started producing copies for "
                f"{self.blueprint.eve_type.name}."
            ),
            user=self.requesting_user,
            level="info",
        )
        other_approvers = set(self.approvers()) - {self.fulfulling_user}
        for approver in other_approvers:
            notify(
                title=(
                    f"Blueprints: {self.blueprint.eve_type.name} request in progress"
                ),
                message=(
                    f"{self.fulfulling_user} has started producing copies for "
                    f"{self.blueprint.eve_type.name} as requested by"
                    f"{self.requesting_user} "
                ),
                user=approver,
                level="info",
            )

    def notify_request_reopened(self, reopened_by: User) -> None:
        """Notify approvers that a blueprint request was reopened."""
        for user in set(self.approvers()) - {reopened_by} | {self.requesting_user}:
            notify(
                title=(f"Blueprints: {self.blueprint.eve_type.name} request re-opened"),
                message=(
                    f"{reopened_by} has re-opened the request for "
                    f"{self.blueprint.eve_type.name} by"
                    f"{self.requesting_user}"
                ),
                user=user,
                level="warning",
            )

    def notify_request_fulfilled(self) -> None:
        """Notify requestor that his blueprint request was fulfilled."""
        notify(
            title=(f"Blueprints: {self.blueprint.eve_type.name} request completed"),
            message=(
                f"{self.fulfulling_user} has finished producing copies for "
                f"{self.blueprint.eve_type.name}."
            ),
            user=self.requesting_user,
            level="success",
        )

    def notify_request_canceled_by_requestor(self) -> None:
        """Notify approvers that a blueprint request was canceled by a requestor."""
        for approver in set(self.approvers()):
            notify(
                title=(f"Blueprints: {self.blueprint.eve_type.name} request canceled"),
                message=(
                    f"{self.requesting_user} has canceled his request for "
                    f"{self.blueprint.eve_type.name}."
                ),
                user=approver,
                level="danger",
            )

    def notify_request_canceled_by_approver(self, canceled_by: User) -> None:
        """Notify approvers that a blueprint request was canceled by an approver."""
        for approver in set(self.approvers()) - {canceled_by} | {self.requesting_user}:
            notify(
                title=(f"Blueprints: {self.blueprint.eve_type.name} request canceled"),
                message=(
                    f"{canceled_by} has canceled the request for "
                    f"{self.blueprint.eve_type.name} "
                    f"by {self.requesting_user}."
                ),
                user=approver,
                level="danger",
            )

    @classmethod
    def approvers(cls) -> models.QuerySet[User]:
        """Return queryset of all approvers."""
        permission = Permission.objects.select_related("content_type").get(
            content_type__app_label=cls._meta.app_label, codename="manage_requests"
        )
        return users_with_permission(permission)



# Generated by Django 4.0.10 on 2023-10-04 12:00

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("blueprints", "0003_query_optimizations"),
    ]

    operations = [
        migrations.AlterField(
            model_name="blueprint",
            name="eve_type",
            field=models.ForeignKey(
                help_text="Blueprint type",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.evetype",
            ),
        ),
        migrations.AlterField(
            model_name="blueprint",
            name="location",
            field=models.ForeignKey(
                help_text="Blueprint location",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blueprints",
                to="blueprints.location",
            ),
        ),
        migrations.AlterField(
            model_name="blueprint",
            name="location_flag",
            field=models.CharField(
                choices=[
                    ("AssetSafety", "Asset Safety"),
                    ("AutoFit", "Auto Fit"),
                    ("Bonus", "Bonus"),
                    ("Booster", "Booster"),
                    ("BoosterBay", "Booster Hold"),
                    ("Capsule", "Capsule"),
                    ("Cargo", "Cargo"),
                    ("CorpDeliveries", "Corp Deliveries"),
                    ("CorpSAG1", "Corp Security Access Group 1"),
                    ("CorpSAG2", "Corp Security Access Group 2"),
                    ("CorpSAG3", "Corp Security Access Group 3"),
                    ("CorpSAG4", "Corp Security Access Group 4"),
                    ("CorpSAG5", "Corp Security Access Group 5"),
                    ("CorpSAG6", "Corp Security Access Group 6"),
                    ("CorpSAG7", "Corp Security Access Group 7"),
                    ("CrateLoot", "Crate Loot"),
                    ("Deliveries", "Deliveries"),
                    ("DroneBay", "Drone Bay"),
                    ("DustBattle", "Dust Battle"),
                    ("DustDatabank", "Dust Databank"),
                    ("FighterBay", "Fighter Bay"),
                    ("FighterTube0", "Fighter Tube 0"),
                    ("FighterTube1", "Fighter Tube 1"),
                    ("FighterTube2", "Fighter Tube 2"),
                    ("FighterTube3", "Fighter Tube 3"),
                    ("FighterTube4", "Fighter Tube 4"),
                    ("FleetHangar", "Fleet Hangar"),
                    ("FrigateEscapeBay", "Frigate escape bay Hangar"),
                    ("Hangar", "Hangar"),
                    ("HangarAll", "Hangar All"),
                    ("HiSlot0", "High power slot 1"),
                    ("HiSlot1", "High power slot 2"),
                    ("HiSlot2", "High power slot 3"),
                    ("HiSlot3", "High power slot 4"),
                    ("HiSlot4", "High power slot 5"),
                    ("HiSlot5", "High power slot 6"),
                    ("HiSlot6", "High power slot 7"),
                    ("HiSlot7", "High power slot 8"),
                    ("HiddenModifiers", "Hidden Modifiers"),
                    ("Implant", "Implant"),
                    ("Impounded", "Impounded"),
                    (
                        "JunkyardReprocessed",
                        "This item was put into a junkyard through reprocessing.",
                    ),
                    (
                        "JunkyardTrashed",
                        "This item was put into a junkyard through being trashed by its owner.",
                    ),
                    ("LoSlot0", "Low power slot 1"),
                    ("LoSlot1", "Low power slot 2"),
                    ("LoSlot2", "Low power slot 3"),
                    ("LoSlot3", "Low power slot 4"),
                    ("LoSlot4", "Low power slot 5"),
                    ("LoSlot5", "Low power slot 6"),
                    ("LoSlot6", "Low power slot 7"),
                    ("LoSlot7", "Low power slot 8"),
                    ("Locked", "Locked item, can not be moved unless unlocked"),
                    ("MedSlot0", "Medium power slot 1"),
                    ("MedSlot1", "Medium power slot 2"),
                    ("MedSlot2", "Medium power slot 3"),
                    ("MedSlot3", "Medium power slot 4"),
                    ("MedSlot4", "Medium power slot 5"),
                    ("MedSlot5", "Medium power slot 6"),
                    ("MedSlot6", "Medium power slot 7"),
                    ("MedSlot7", "Medium power slot 8"),
                    ("OfficeFolder", "Office Folder"),
                    ("Pilot", "Pilot"),
                    ("PlanetSurface", "Planet Surface"),
                    ("QuafeBay", "Quafe Bay"),
                    ("QuantumCoreRoom", "Quantum Core Room"),
                    ("Reward", "Reward"),
                    ("RigSlot0", "Rig power slot 1"),
                    ("RigSlot1", "Rig power slot 2"),
                    ("RigSlot2", "Rig power slot 3"),
                    ("RigSlot3", "Rig power slot 4"),
                    ("RigSlot4", "Rig power slot 5"),
                    ("RigSlot5", "Rig power slot 6"),
                    ("RigSlot6", "Rig power slot 7"),
                    ("RigSlot7", "Rig power slot 8"),
                    ("SecondaryStorage", "Secondary Storage"),
                    ("ServiceSlot0", "Service Slot 0"),
                    ("ServiceSlot1", "Service Slot 1"),
                    ("ServiceSlot2", "Service Slot 2"),
                    ("ServiceSlot3", "Service Slot 3"),
                    ("ServiceSlot4", "Service Slot 4"),
                    ("ServiceSlot5", "Service Slot 5"),
                    ("ServiceSlot6", "Service Slot 6"),
                    ("ServiceSlot7", "Service Slot 7"),
                    ("ShipHangar", "Ship Hangar"),
                    ("ShipOffline", "Ship Offline"),
                    ("Skill", "Skill"),
                    ("SkillInTraining", "Skill In Training"),
                    ("SpecializedAmmoHold", "Specialized Ammo Hold"),
                    ("SpecializedCommandCenterHold", "Specialized Command Center Hold"),
                    ("SpecializedFuelBay", "Specialized Fuel Bay"),
                    ("SpecializedGasHold", "Specialized Gas Hold"),
                    (
                        "SpecializedIndustrialShipHold",
                        "Specialized Industrial Ship Hold",
                    ),
                    ("SpecializedLargeShipHold", "Specialized Large Ship Hold"),
                    ("SpecializedMaterialBay", "Specialized Material Bay"),
                    ("SpecializedMediumShipHold", "Specialized Medium Ship Hold"),
                    ("SpecializedMineralHold", "Specialized Mineral Hold"),
                    ("SpecializedOreHold", "Specialized Ore Hold"),
                    (
                        "SpecializedPlanetaryCommoditiesHold",
                        "Specialized Planetary Commodities Hold",
                    ),
                    ("SpecializedSalvageHold", "Specialized Salvage Hold"),
                    ("SpecializedShipHold", "Specialized Ship Hold"),
                    ("SpecializedSmallShipHold", "Specialized Small Ship Hold"),
                    ("StructureActive", "Structure Active"),
                    ("StructureFuel", "Structure Fuel"),
                    ("StructureInactive", "Structure Inactive"),
                    ("StructureOffline", "Structure Offline"),
                    ("SubSystemBay", "Sub System Bay"),
                    ("SubSystemSlot0", "Sub System Slot 0"),
                    ("SubSystemSlot1", "Sub System Slot 1"),
                    ("SubSystemSlot2", "Sub System Slot 2"),
                    ("SubSystemSlot3", "Sub System Slot 3"),
                    ("SubSystemSlot4", "Sub System Slot 4"),
                    ("SubSystemSlot5", "Sub System Slot 5"),
                    ("SubSystemSlot6", "Sub System Slot 6"),
                    ("SubSystemSlot7", "Sub System Slot 7"),
                    ("Unlocked", "Unlocked item, can be moved"),
                    ("Wallet", "Wallet"),
                    ("Wardrobe", "Wardrobe"),
                    ("Undefined", "undefined"),
                ],
                help_text="Additional location information",
                max_length=36,
            ),
        ),
        migrations.AlterField(
            model_name="blueprint",
            name="owner",
            field=models.ForeignKey(
                help_text="Corporation that owns the blueprint",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blueprints",
                to="blueprints.owner",
            ),
        ),
    ]



"""Decorators for Blueprints."""

from functools import wraps

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


# TODO: This is currently not used and can potentially be removed
def fetch_token_for_owner(scopes):
    """returns valid token for owner.
    Needs to be attached on an Owner method !!

    Args:
    -scopes: Provide the required scopes.
    """

    def decorator(func):
        @wraps(func)
        def _wrapped_view(owner, *args, **kwargs):
            token = owner.valid_token(scopes)
            logger.debug(
                "%s: Using token %s for `%s`",
                token.character_name,
                token.pk,
                func.__name__,
            )
            return func(owner, token, *args, **kwargs)

        return _wrapped_view

    return decorator



"""Tasks for Blueprints."""

from celery import Task, shared_task

from esi.models import Token

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from app_utils.esi import retry_task_on_esi_issue
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import BLUEPRINTS_TASKS_TIME_LIMIT
from .models import Location, Owner

DEFAULT_TASK_PRIORITY = 6
logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@shared_task(time_limit=BLUEPRINTS_TASKS_TIME_LIMIT)
def update_all_blueprints():
    """Update all blueprints."""
    for owner in Owner.objects.filter(is_active=True):
        update_blueprints_for_owner.apply_async(
            kwargs={"owner_pk": owner.pk}, priority=DEFAULT_TASK_PRIORITY
        )


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"keys": ["owner_pk"], "graceful": True},
    time_limit=BLUEPRINTS_TASKS_TIME_LIMIT,
)
def update_blueprints_for_owner(self: Task, owner_pk: int):
    """Fetch all blueprints for an owner from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    with retry_task_on_esi_issue(self):
        owner.update_blueprints_esi()


@shared_task(time_limit=BLUEPRINTS_TASKS_TIME_LIMIT)
def update_all_industry_jobs():
    """Update all industry jobs."""
    for owner in Owner.objects.filter(is_active=True):
        update_industry_jobs_for_owner.apply_async(
            kwargs={"owner_pk": owner.pk}, priority=DEFAULT_TASK_PRIORITY
        )


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"keys": ["owner_pk"], "graceful": True},
    time_limit=BLUEPRINTS_TASKS_TIME_LIMIT,
)
def update_industry_jobs_for_owner(self: Task, owner_pk: int):
    """Fetch all industry jobs for an owner from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    with retry_task_on_esi_issue(self):
        owner.update_industry_jobs_esi()


@shared_task(time_limit=BLUEPRINTS_TASKS_TIME_LIMIT)
def update_all_locations():
    """Update all locations."""
    for owner in Owner.objects.filter(is_active=True):
        update_locations_for_owner.apply_async(
            kwargs={"owner_pk": owner.pk}, priority=DEFAULT_TASK_PRIORITY
        )


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"keys": ["owner_pk"], "graceful": True},
    time_limit=BLUEPRINTS_TASKS_TIME_LIMIT,
)
def update_locations_for_owner(self: Task, owner_pk: int):
    """Fetch all blueprints for an owner from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    with retry_task_on_esi_issue(self):
        owner.update_locations_esi()


@shared_task(
    bind=True,
    max_retries=None,
    base=QueueOnce,
    once={"keys": ["structure_id"], "graceful": True},
    time_limit=BLUEPRINTS_TASKS_TIME_LIMIT,
)
def update_structure_esi(self, structure_id: int, token_pk: int):
    """Update a structure object from ESI."""
    token = Token.objects.get(pk=token_pk)
    with retry_task_on_esi_issue(self):
        Location.objects.structure_update_or_create_esi(id=structure_id, token=token)



from unittest.mock import Mock, patch

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from eveuniverse.models import EveEntity, EveType

from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import create_user_from_evecharacter, json_response_to_python

from blueprints.models import Blueprint, Location, Owner, Request
from blueprints.views.blueprint_list import BlueprintListJson
from blueprints.views.regular_views import (
    add_corporation_blueprint_owner,
    add_personal_blueprint_owner,
    list_blueprints_ffd,
    list_user_owners,
    remove_owner,
)

from . import create_owner
from .testdata.load_entities import load_entities
from .testdata.load_eveuniverse import load_eveuniverse
from .testdata.load_locations import load_locations

MODELS_PATH = "blueprints.models"
VIEWS_PATH = "blueprints.views.regular_views"


def notification_count(user) -> int:
    return user.notification_set.filter(viewed=False).count()


class _TestCaseWithTestData(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.owner = create_owner(character_id=1101, corporation_id=2101)
        cls.user_owner = cls.owner.character.user
        cls.corporation_2001 = EveEntity.objects.get(id=2101)
        cls.jita_44 = Location.objects.get(id=60003760)
        cls.blueprint = Blueprint.objects.create(
            location=cls.jita_44,
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner,
            runs=None,
            quantity=1,
            location_flag="AssetSafety",
            material_efficiency=0,
            time_efficiency=0,
            item_id=1,
        )
        cls.user_requestor, _ = create_user_from_evecharacter(
            1102,
            permissions=["blueprints.basic_access", "blueprints.request_blueprints"],
        )
        cls.user_other_approver, _ = create_user_from_evecharacter(
            1103,
            permissions=["blueprints.basic_access", "blueprints.manage_requests"],
        )


class TestBlueprintsData(_TestCaseWithTestData):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.user_owner = AuthUtils.add_permission_to_user_by_name(
            "blueprints.view_blueprint_locations", cls.user_owner
        )

    def test_blueprints_data(self):
        request = self.factory.get(reverse("blueprints:list_blueprints"))
        request.user = self.user_owner
        response = BlueprintListJson.as_view()(request)
        self.assertEqual(response.status_code, 200)
        data = json_response_to_python(response)["data"]
        self.assertEqual(len(data), 1)
        row = data[0]
        self.assertEqual(row[1], "Mobile Tractor Unit Blueprint")
        self.assertEqual(row[9], "Jita IV - Moon 4 - Caldari Navy Assembly Plant")

    def test_my_requests_data(self):
        Request.objects.create(
            blueprint=self.blueprint,
            runs=None,
            requesting_user=self.user_owner,
            fulfulling_user=None,
            status="OP",
        )
        request = self.factory.get(reverse("blueprints:list_user_requests"))
        request.user = self.user_owner
        response = BlueprintListJson.as_view()(request)
        self.assertEqual(response.status_code, 200)
        data = json_response_to_python(response)["data"]
        self.assertEqual(len(data), 1)
        row = data[0]
        self.assertEqual(row[1], "Mobile Tractor Unit Blueprint")
        self.assertEqual(row[9], "Jita IV - Moon 4 - Caldari Navy Assembly Plant")

    def test_list_user_owners(self):
        request = self.factory.get(reverse("blueprints:list_user_owners"))
        request.user = self.user_owner
        response = list_user_owners(request)
        self.assertEqual(response.status_code, 200)
        data = json_response_to_python(response)
        self.assertEqual(len(data), 1)
        row = data[0]
        self.assertEqual(row["name"], "Lexcorp")
        self.assertEqual(row["type"], "corporate")

    @patch(VIEWS_PATH + ".messages")
    def test_remove_owner(self, mock_messages):
        request = self.factory.post(
            reverse("blueprints:remove_owner", args=[self.owner.pk])
        )
        request.user = self.user_owner
        response = remove_owner(request, self.owner.pk)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.info.called)
        self.assertEqual(Owner.objects.filter(pk=self.owner.pk).first(), None)

    def test_should_handle_owner_without_character(self):
        # given
        Owner.objects.create(
            corporation=EveCorporationInfo.objects.get(corporation_id=2001)
        )  # owner without character
        request = self.factory.get(reverse("blueprints:list_blueprints"))
        request.user = self.user_owner
        # when
        response = BlueprintListJson.as_view()(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_python(response)["data"]
        self.assertEqual(len(data), 1)
        row = data[0]
        self.assertEqual(row[1], "Mobile Tractor Unit Blueprint")
        self.assertEqual(row[9], "Jita IV - Moon 4 - Caldari Navy Assembly Plant")

    def test_should_handle_empty_owner(self):
        # given
        Owner.objects.create()  # empty owner
        request = self.factory.get(reverse("blueprints:list_blueprints"))
        request.user = self.user_owner
        # when
        response = BlueprintListJson.as_view()(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_python(response)["data"]
        self.assertEqual(len(data), 1)
        row = data[0]
        self.assertEqual(row[1], "Mobile Tractor Unit Blueprint")
        self.assertEqual(row[9], "Jita IV - Moon 4 - Caldari Navy Assembly Plant")


class TestListBlueprintsFdd(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.factory = RequestFactory()
        cls.owner_1001 = create_owner(character_id=1001, corporation_id=None)
        cls.owner_1001.character.user = AuthUtils.add_permission_to_user_by_name(
            "blueprints.view_blueprint_locations", cls.owner_1001.character.user
        )
        cls.owner_1002 = create_owner(character_id=1002, corporation_id=2001)

    def test_should_return_list_of_options(self):
        # given
        Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=self.owner_1001,
            runs=10,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=1,
        )
        Blueprint.objects.create(
            location=Location.objects.get(id=1000000000001),
            eve_type=EveType.objects.get(id=33519),
            owner=self.owner_1002,
            location_flag="AssetSafety",
            material_efficiency=20,
            time_efficiency=40,
            item_id=2,
        )
        request = self.factory.get(
            reverse("blueprints:list_blueprints_ffd")
            + "?columns=location,owner,material_efficiency,time_efficiency,is_original"
        )
        request.user = self.owner_1001.character.user
        # when
        response = list_blueprints_ffd(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_python(response)
        self.assertDictEqual(
            data,
            {
                "location": [
                    "Amamake - Test Structure Alpha",
                    "Jita IV - Moon 4 - Caldari Navy Assembly Plant",
                ],
                "owner": ["Bruce Wayne", "Wayne Technologies"],
                "material_efficiency": [10, 20],
                "time_efficiency": [30, 40],
                "is_original": ["no", "yes"],
            },
        )


class TestRequestWorkflow(_TestCaseWithTestData):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

    @patch(VIEWS_PATH + ".messages")
    def test_should_create_new_request(self, mock_messages):
        # given
        self.client.force_login(self.user_requestor)
        # when
        response = self.client.post(
            "/blueprints/requests/add", data={"pk": self.blueprint.pk}
        )
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.info.called)
        user_request = Request.objects.filter(blueprint=self.blueprint).first()
        self.assertEqual(user_request.status, Request.STATUS_OPEN)
        self.assertEqual(user_request.requesting_user, self.user_requestor)
        self.assertEqual(notification_count(self.user_requestor), 0)
        self.assertEqual(notification_count(self.user_owner), 1)
        self.assertEqual(notification_count(self.user_other_approver), 1)

    @patch(VIEWS_PATH + ".messages")
    def test_should_mark_request_cancelled_by_requestor(self, mock_messages):
        # given
        user_request = Request.objects.create(
            blueprint=self.blueprint,
            runs=None,
            requesting_user=self.user_requestor,
            status=Request.STATUS_IN_PROGRESS,
        )
        self.client.force_login(self.user_requestor)
        # when
        response = self.client.post(f"/blueprints/requests/{user_request.pk}/cancel")
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.info.called)
        user_request.refresh_from_db()
        self.assertEqual(user_request.status, Request.STATUS_CANCELLED)
        self.assertIsNone(user_request.fulfulling_user)
        self.assertEqual(notification_count(self.user_requestor), 0)
        self.assertEqual(notification_count(self.user_owner), 1)
        self.assertEqual(notification_count(self.user_other_approver), 1)

    @patch(VIEWS_PATH + ".messages")
    def test_should_mark_request_as_in_progress(self, mock_messages):
        # given
        user_request = Request.objects.create(
            blueprint=self.blueprint,
            runs=None,
            requesting_user=self.user_requestor,
            status=Request.STATUS_OPEN,
        )
        self.client.force_login(self.user_owner)
        # when
        response = self.client.post(
            f"/blueprints/requests/{user_request.pk}/in_progress"
        )
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.info.called)
        user_request.refresh_from_db()
        self.assertEqual(user_request.status, Request.STATUS_IN_PROGRESS)
        self.assertEqual(notification_count(self.user_requestor), 1)
        self.assertEqual(notification_count(self.user_owner), 0)
        self.assertEqual(notification_count(self.user_other_approver), 1)

    @patch(VIEWS_PATH + ".messages")
    def test_should_mark_request_as_fulfilled(self, mock_messages):
        # given
        user_request = Request.objects.create(
            blueprint=self.blueprint,
            runs=None,
            requesting_user=self.user_requestor,
            status=Request.STATUS_IN_PROGRESS,
        )
        self.client.force_login(self.user_owner)
        # when
        response = self.client.post(f"/blueprints/requests/{user_request.pk}/fulfill")
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.info.called)
        user_request.refresh_from_db()
        self.assertEqual(user_request.status, Request.STATUS_FULFILLED)
        self.assertEqual(user_request.fulfulling_user, self.user_owner)
        self.assertEqual(notification_count(self.user_requestor), 1)
        self.assertEqual(notification_count(self.user_owner), 0)
        self.assertEqual(notification_count(self.user_other_approver), 0)

    @patch(VIEWS_PATH + ".messages")
    def test_should_mark_request_as_cancelled_by_owner(self, mock_messages):
        # given
        user_request = Request.objects.create(
            blueprint=self.blueprint,
            runs=None,
            requesting_user=self.user_requestor,
            status=Request.STATUS_IN_PROGRESS,
        )
        self.client.force_login(self.user_owner)
        # when
        response = self.client.post(f"/blueprints/requests/{user_request.pk}/cancel")
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.info.called)
        user_request.refresh_from_db()
        self.assertEqual(user_request.status, Request.STATUS_CANCELLED)
        self.assertIsNone(user_request.fulfulling_user)
        self.assertEqual(notification_count(self.user_requestor), 1)
        self.assertEqual(notification_count(self.user_owner), 0)
        self.assertEqual(notification_count(self.user_other_approver), 1)

    @patch(VIEWS_PATH + ".messages")
    def test_should_mark_request_as_reopened(self, mock_messages):
        # given
        user_request = Request.objects.create(
            blueprint=self.blueprint,
            runs=None,
            requesting_user=self.user_requestor,
            fulfulling_user=self.user_owner,
            status=Request.STATUS_IN_PROGRESS,
        )
        self.client.force_login(self.user_owner)
        # when
        response = self.client.post(f"/blueprints/requests/{user_request.pk}/open")
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.info.called)
        user_request.refresh_from_db()
        self.assertEqual(user_request.status, Request.STATUS_OPEN)
        self.assertIsNone(user_request.fulfulling_user)
        self.assertEqual(notification_count(self.user_requestor), 1)
        self.assertEqual(notification_count(self.user_owner), 0)
        self.assertEqual(notification_count(self.user_other_approver), 1)


class TestOtherViews(_TestCaseWithTestData):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

    def test_can_open_index_by_requestor(self):
        # given
        self.client.force_login(self.user_requestor)
        # when
        response = self.client.get("/blueprints/")
        # then
        self.assertTemplateUsed(response, "blueprints/index.html")

    def test_can_open_index_by_owner(self):
        # given
        self.client.force_login(self.user_owner)
        # when
        response = self.client.get("/blueprints/")
        # then
        self.assertTemplateUsed(response, "blueprints/index.html")

    def test_can_open_blueprint_modal(self):
        # given
        self.client.force_login(self.user_owner)
        # when
        response = self.client.get(
            f"/blueprints/modals/view_blueprint?blueprint_id={self.blueprint.pk}"
        )
        # then
        self.assertContains(response, self.blueprint.eve_type.name)

    def test_can_open_request_modal(self):
        # given
        user_request = Request.objects.create(
            blueprint=self.blueprint,
            runs=None,
            requesting_user=self.user_requestor,
            status=Request.STATUS_IN_PROGRESS,
        )
        self.client.force_login(self.user_owner)
        # when
        response = self.client.get(
            f"/blueprints/modals/view_request?request_id={user_request.pk}"
        )
        # then
        self.assertContains(response, self.blueprint.eve_type.name)


@patch(VIEWS_PATH + ".tasks.update_locations_for_owner")
@patch(VIEWS_PATH + ".tasks.update_blueprints_for_owner")
@patch(VIEWS_PATH + ".notify_admins")
@patch(VIEWS_PATH + ".messages")
class TestAddCorporationBlueprintOwner(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.user, _ = create_user_from_evecharacter(
            1001,
            permissions=[
                "blueprints.basic_access",
                "blueprints.add_corporate_blueprint_owner",
            ],
        )

    def _add_corporate_blueprint_owner(self, token=None, user=None):
        # given
        request = self.factory.get(reverse("blueprints:add_corporate_blueprint_owner"))
        if not user:
            user = self.user
        if not token:
            token = user.token_set.first()
        request.user = user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = add_corporation_blueprint_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        return orig_view(request, token)

    @patch(VIEWS_PATH + ".BLUEPRINTS_ADMIN_NOTIFICATIONS_ENABLED", True)
    def test_should_add_new_owner_and_notify_admins(
        self,
        mock_messages,
        mock_notify_admins,
        mock_update_blueprints_for_owner,
        mock_update_locations_for_owner,
    ):
        # when
        response = self._add_corporate_blueprint_owner()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("blueprints:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertTrue(mock_notify_admins.called)
        owner = Owner.objects.first()
        self.assertEqual(
            owner.corporation.corporation_id,
            self.user.profile.main_character.corporation_id,
        )
        self.assertTrue(mock_update_blueprints_for_owner.delay.called)
        self.assertTrue(mock_update_locations_for_owner.delay.called)

    @patch(VIEWS_PATH + ".BLUEPRINTS_ADMIN_NOTIFICATIONS_ENABLED", False)
    def test_should_add_new_owner_and_not_notify_admins(
        self,
        mock_messages,
        mock_notify_admins,
        mock_update_blueprints_for_owner,
        mock_update_locations_for_owner,
    ):
        # when
        response = self._add_corporate_blueprint_owner()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("blueprints:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertFalse(mock_notify_admins.called)
        owner = Owner.objects.first()
        self.assertEqual(
            owner.corporation.corporation_id,
            self.user.profile.main_character.corporation_id,
        )
        self.assertTrue(mock_update_blueprints_for_owner.delay.called)
        self.assertTrue(mock_update_locations_for_owner.delay.called)


@patch(VIEWS_PATH + ".tasks.update_locations_for_owner")
@patch(VIEWS_PATH + ".tasks.update_blueprints_for_owner")
@patch(VIEWS_PATH + ".notify_admins")
@patch(VIEWS_PATH + ".messages")
class TestAddPersonalBlueprintOwner(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.user, _ = create_user_from_evecharacter(
            1001,
            permissions=[
                "blueprints.basic_access",
                "blueprints.add_personal_blueprint_owner",
            ],
        )

    def _add_personal_blueprint_owner(self, token=None, user=None):
        # given
        request = self.factory.get(reverse("blueprints:add_personal_blueprint_owner"))
        if not user:
            user = self.user
        if not token:
            token = user.token_set.first()
        request.user = user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = add_personal_blueprint_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        return orig_view(request, token)

    @patch(VIEWS_PATH + ".BLUEPRINTS_ADMIN_NOTIFICATIONS_ENABLED", True)
    def test_should_add_new_owner_and_notify_admins(
        self,
        mock_messages,
        mock_notify_admins,
        mock_update_blueprints_for_owner,
        mock_update_locations_for_owner,
    ):
        # when
        response = self._add_personal_blueprint_owner()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("blueprints:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertTrue(mock_notify_admins.called)
        owner = Owner.objects.first()
        self.assertIsNone(owner.corporation)
        self.assertEqual(
            owner.character.character.character_id,
            self.user.profile.main_character.character_id,
        )
        self.assertTrue(mock_update_blueprints_for_owner.delay.called)
        self.assertTrue(mock_update_locations_for_owner.delay.called)
        _, kwargs = mock_notify_admins.call_args
        self.assertIn(owner.character.character.character_name, kwargs["title"])
        self.assertIn(owner.character.character.character_name, kwargs["message"])

    @patch(VIEWS_PATH + ".BLUEPRINTS_ADMIN_NOTIFICATIONS_ENABLED", False)
    def test_should_add_new_owner_and_not_notify_admins(
        self,
        mock_messages,
        mock_notify_admins,
        mock_update_blueprints_for_owner,
        mock_update_locations_for_owner,
    ):
        # when
        response = self._add_personal_blueprint_owner()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("blueprints:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertFalse(mock_notify_admins.called)
        owner = Owner.objects.first()
        self.assertIsNone(owner.corporation)
        self.assertEqual(
            owner.character.character.character_id,
            self.user.profile.main_character.character_id,
        )
        self.assertTrue(mock_update_blueprints_for_owner.delay.called)
        self.assertTrue(mock_update_locations_for_owner.delay.called)
